package cholesky

import java.io._
import edu.rice.cnc.api._


/**
 * @author Shams Imam (shams@rice.edu)
 * @since 1/2/12 10:34 AM
 */

object CholeskyMain {

  //     Read the matrix from the input file. The numbers in the matrix
  //   are whitespace separated
  def matrix_init(A: Array[Array[Double]], n: Int, fname: String): Unit = {
    val r: Reader = new BufferedReader(new FileReader(fname))
    val stok: StreamTokenizer = new StreamTokenizer(r)
    stok.parseNumbers()
    stok.nextToken()
    var i: Int = 0
    while (i < n) {
      var j: Int = 0
      while (j < n) {
        if (stok.ttype == StreamTokenizer.TT_EOF) {
          throw new EOFException("Matrix size incorrect")
        }
        if (stok.ttype != StreamTokenizer.TT_NUMBER) {
          throw new IOException("A non-number has been read from the file: " + stok.sval)
        }
        A(i)(j) = stok.nval
        stok.nextToken()
        j += 1;
      }
      i += 1;
    }
  }


  def main(args: Array[String]): Unit = {
    var n: Int = 0
    var b: Int = 0
    var filename: String = null
    var A: Array[Array[Double]] = null

    if (args.length == 3) {
      n = Integer.parseInt(args(0))
      b = Integer.parseInt(args(1))
      filename = args(2)
    } else {
      println("Usage: cholesky matrixSize tileSize fileName")
      return
    }

    if (n % b != 0) {
      println("Incorrect tile size " + b + " for the matrix of size " + n)
      return
    }

    A = Array.ofDim[Array[Double]](n)
    var i = 0
    while (i < n) {
      A(i) = Array.ofDim[Double](n)
      i += 1
    }

    val s1ComputeStep = new UserS1ComputeStepStep()
    val s2ComputeStep = new UserS2ComputeStepStep()
    val s3ComputeStep = new UserS3ComputeStepStep()
    val kComputeStep = new UserKComputeStepStep()
    val kjComputeStep = new UserKjComputeStepStep()
    val kjiComputeStep = new UserKjiComputeStepStep()

    val graph = new CholeskyGraph(
      kComputeStep, kjComputeStep, s1ComputeStep,
      kjiComputeStep, s2ComputeStep, s3ComputeStep)

    val numTiles: Int = n / b
    val tileSize: Int = b
    println("CholeskyMain: num tiles = " + numTiles + ", tile size = " + tileSize)
    matrix_init(A, n, filename)

    val begin: Long = System.currentTimeMillis()
    graph.run(new Runnable() {
      def run(): Unit = {

        graph.numTiles.put(Point.ZERO_1D, numTiles)
        graph.tileSize.put(Point.ZERO_1D, tileSize)

        var A_i: Int = 0
        var A_j: Int = 0
        var T_i: Int = 0
        var T_j: Int = 0
        var i: Int = 0

        while (i < numTiles) {

          var j: Int = 0
          while (j <= i) {

            val temp = Array.ofDim[Double](tileSize, tileSize)
            A_i = i * b
            T_i = 0
            while (T_i < tileSize) {
              A_j = j * b
              T_j = 0
              while (T_j < tileSize) {
                temp(T_i)(T_j) = A(A_i)(A_j)
                A_j += 1;
                T_j += 1;
              }
              A_i += 1;
              T_i += 1;
            }

            val loopTag: Point = Point.create(i, j, 0)
            graph.Lkji.put(loopTag, temp)

            j += 1;
          }

          i += 1;
        }

        graph.singleton.put(Point.ZERO_1D)
      }
    })

    val end: Long = System.currentTimeMillis()
    val execTime: Long = end - begin
    System.out.println("Execution Time: " + execTime + " ms.")

    System.out.println("Lkji.size = " + graph.Lkji.size)
    graph.Lkji.dump(System.out)

    var k: Int = 0
    val w: BufferedWriter = new BufferedWriter(new FileWriter("cholesky_out.txt"))
    var ni: Int = 0
    while (ni < numTiles) {
      var i_b: Int = 0
      while (i_b < tileSize) {
        k = 1
        var j: Int = 0
        while (j <= ni) {
          val loopTag: Point = Point.create(ni, j, k)
          val mat_out: Array[Array[Double]] = graph.Lkji.safeGet(loopTag)
          if (ni != j) {
            var j_b: Int = 0
            while (j_b < tileSize) {
              w.write(mat_out(i_b)(j_b) + " ")
              j_b += 1;
            }
          }
          else {
            var j_b: Int = 0
            while (j_b <= i_b) {
              w.write(mat_out(i_b)(j_b) + " ")
              j_b += 1;
            }
          }
          k += 1;
          j += 1;
        }
        i_b += 1;
      }
      ni += 1;
    }

    w.flush()
    w.close()
  }

}