package cholesky

import edu.rice.cnc.api._
import util.continuations.cps


/**
 * @author Shams Imam (shams@rice.edu)
 * @since 1/2/12 19:35 PM
 */

class UserKComputeStepStep extends KComputeStepStep {
  def compute(
    tag: Point,
    inNumTiles: InputCollection[Point, Int],
    outControlS1: TagCollection[Point]
    ): Unit@cps[Any] = {

    val numTiles = inNumTiles.get(Point.ZERO_1D)

    var k = 0
    while (k < numTiles) {
      outControlS1.put(Point.create(k))
      k += 1
    }

  }
}

class UserKjComputeStepStep extends KjComputeStepStep {
  def compute(
    tag: Point,
    inNumTiles: InputCollection[Point, Int],
    outControlS2: TagCollection[Point]
    ): Unit@cps[Any] = {

    val numTiles = inNumTiles.get(Point.ZERO_1D)
    val k = tag(0)

    var j = k + 1
    while (j < numTiles) {
      outControlS2.put(Point.create(k, j))
      j += 1
    }
  }
}

class UserKjiComputeStepStep extends KjiComputeStepStep {
  def compute(
    tag: Point,
    outControlS3: TagCollection[Point]
    ): Unit@cps[Any] = {

    val k = tag(0)
    val j = tag(1)

    var i = k + 1
    while (i <= j) {
      outControlS3.put(Point.create(k, j, i))
      i += 1
    }
  }
}

class UserS1ComputeStepStep extends S1ComputeStepStep {
  def compute(
    tag: Point,
    inTileSize: InputCollection[Point, Int],
    inLkji: InputCollection[Point, Array[Array[Double]]],
    outLkji: OutputCollection[Point, Array[Array[Double]]]
    ): Unit@cps[Any] = {

    val k = tag(0)
    val tileSize = inTileSize.get(Point.ZERO_1D)
    val kkk: Point = Point.create(k, k, k)
    val aBlock = inLkji.get(kkk)

    // val lBlock = Array.ofDim[Double](tileSize, tileSize)
    val lBlock = Array.ofDim[Array[Double]](tileSize)
    var i = 0
    while (i < tileSize) {
      lBlock(i) = Array.ofDim[Double](i + 1)
      i += 1
    }

    var kB = 0
    while (kB < tileSize) {
      if (aBlock(kB)(kB) <= 0) {
        throw new IllegalArgumentException("Not a symmetric positive definite (SPD) matrix")
      }

      lBlock(kB)(kB) = scala.math.sqrt(aBlock(kB)(kB))

      var jB = kB + 1
      while (jB < tileSize) {
        lBlock(jB)(kB) = aBlock(jB)(kB) / lBlock(kB)(kB)
        jB += 1;
      }

      var jBB = kB + 1
      while (jBB < tileSize) {
        var iB = jBB
        while (iB < tileSize) {
          aBlock(iB)(jBB) = aBlock(iB)(jBB) - (lBlock(iB)(kB) * lBlock(jBB)(kB))
          iB += 1;
        }
        jBB += 1;
      }
      kB += 1;
    }

    val newTag: Point = Point.create(k, k, k + 1)
    outLkji.put(newTag, lBlock)
  }
}

class UserS2ComputeStepStep extends S2ComputeStepStep {
  def compute(
    tag: Point,
    inTileSize: InputCollection[Point, Int],
    inLkji: InputCollection[Point, Array[Array[Double]]],
    outLkji: OutputCollection[Point, Array[Array[Double]]]
    ): Unit@cps[Any] = {

    val k = tag(0)
    val j = tag(1)

    // get the input tile.
    val aBlock = inLkji.get(Point.create(j, k, k))

    // get the 2nd input tile (Output of previous step).
    val liBlock = inLkji.get(Point.create(k, k, k + 1));

    val tileSize = inTileSize.get(Point.ZERO_1D)

    val loBlock = Array.ofDim[Double](tileSize, tileSize)

    var kB = 0
    while (kB < tileSize) {

      var iB = 0
      while (iB < tileSize) {
        loBlock(iB)(kB) = aBlock(iB)(kB) / liBlock(kB)(kB)
        iB += 1
      }

      var jB = kB + 1
      while (jB < tileSize) {
        var iB = 0
        while (iB < tileSize) {
          aBlock(iB)(jB) -= liBlock(jB)(kB) * loBlock(iB)(kB)
          iB += 1;
        }
        jB += 1
      }
      kB += 1;
    }

    // Write the output tile at the next time step.
    outLkji.put(Point.create(j, k, k + 1), loBlock);
  }

}


class UserS3ComputeStepStep extends S3ComputeStepStep {
  def compute(
    tag: Point,
    inTileSize: InputCollection[Point, Int],
    inLkji: InputCollection[Point, Array[Array[Double]]],
    outLkji: OutputCollection[Point, Array[Array[Double]]]
    ): Unit@cps[Any] = {

    val j = tag(1)
    val i = tag(2)

    if (i == j) {
      computeDiagonal(tag, inTileSize, inLkji, outLkji)
    } else {
      computeNonDiagonal(tag, inTileSize, inLkji, outLkji)
    }
  }

  private def computeDiagonal(
    tag: Point,
    inTileSize: InputCollection[Point, Int],
    inLkji: InputCollection[Point, Array[Array[Double]]],
    outLkji: OutputCollection[Point, Array[Array[Double]]]
    ): Unit@cps[Any] = {

    val k = tag(0)
    val j = tag(1)
    val i = tag(2)

    val aBlock = inLkji.get(Point.create(j, i, k))
    // In case of a diagonal tile, i=j, hence both the tiles are the same.
    val l2Block = inLkji.get(Point.create(j, k, k + 1))

    val tileSize = inTileSize.get(Point.ZERO_1D)

    var jB = 0
    while (jB < tileSize) {

      var kB = 0
      while (kB < tileSize) {
        val temp: Double = 0 - l2Block(jB)(kB)

        var iB = jB
        while (iB < tileSize) {
          aBlock(iB)(jB) += temp * l2Block(iB)(kB)
          iB += 1;
        }

        kB += 1;
      }

      jB += 1
    }

    // Write the output at the next time step.
    outLkji.put(Point.create(j, i, k + 1), aBlock)
  }

  private def computeNonDiagonal(
    tag: Point,
    inTileSize: InputCollection[Point, Int],
    inLkji: InputCollection[Point, Array[Array[Double]]],
    outLkji: OutputCollection[Point, Array[Array[Double]]]
    ): Unit@cps[Any] = {

    val k = tag(0)
    val j = tag(1)
    val i = tag(2)

    val aBlock = inLkji.get(Point.create(j, i, k))

    // Non-diagonal tile, get the second tile.
    val l1Block = inLkji.get(Point.create(j, k, k + 1))

    // Non-diagonal tile, get the first tile.
    val l2Block = inLkji.get(Point.create(i, k, k + 1))

    val tileSize = inTileSize.get(Point.ZERO_1D)

    var jB = 0
    while (jB < tileSize) {

      var kB = 0
      while (kB < tileSize) {
        val temp: Double = 0 - l2Block(jB)(kB)

        var iB = 0
        while (iB < tileSize) {
          aBlock(iB)(jB) += temp * l1Block(iB)(kB)
          iB += 1;
        }

        kB += 1;
      }

      jB += 1
    }

    // Write the output at the next time step.
    outLkji.put(Point.create(j, i, k + 1), aBlock)
  }
}