package ludecomp

import edu.rice.cnc.api._
import java.io._

object LUDecompositionMain {
  def main(args: Array[String]): Unit = {

    var n: Int = 0
    var b: Int = 0
    var verbose: Boolean = false
    var GC: Boolean = false
    var filename: String = null
    var A: Array[Array[Double]] = null

    if (args.length == 4) {
      GC = (args(0).equalsIgnoreCase("-GC"))
      n = Integer.parseInt(args(1))
      b = Integer.parseInt(args(2))
      filename = args(3)
    } else if (args.length == 5 && args(1).equalsIgnoreCase("-dumpOutput")) {
      GC = (args(0).equalsIgnoreCase("-GC"))
      if (false) System.err.println("Using the verbose mode")
      n = Integer.parseInt(args(2))
      b = Integer.parseInt(args(3))
      filename = args(4)
      verbose = true
    } else {
      println("Usage: lu {-GC | -noGC} [-dumpOutput] matrixSize tileSize fileName")
      return
    }

    if (n % b != 0) {
      println("Incorrect tile size " + b + " for the matrix of size " + n)
      return
    }

    A = Array.ofDim[Double](n, n)

    try {
      matrix_init(A, n, filename)

      val columnPeripheryStep: UserColumnPeripheryStepStep = new UserColumnPeripheryStepStep()
      val subMatrixStep: UserSubMatrixStepStep = new UserSubMatrixStepStep()
      val createColumnPeripheryTagsStep: UserCreateColumnPeripheryTagsStepStep = new UserCreateColumnPeripheryTagsStepStep()
      val createSubMatrixTagsStep: UserCreateSubMatrixTagsStepStep = new UserCreateSubMatrixTagsStepStep()
      val createRowPeripheryTagsStep: UserCreateRowPeripheryTagsStepStep = new UserCreateRowPeripheryTagsStepStep()
      val pivotStep: UserPivotStepStep = new UserPivotStepStep()
      val createPivotTagsStep: UserCreatePivotTagsStepStep = new UserCreatePivotTagsStepStep()
      val rowPeripheryStep: UserRowPeripheryStepStep = new UserRowPeripheryStepStep()

      val graph: LUDecompositionGraph = new LUDecompositionGraph(
        createPivotTagsStep, createColumnPeripheryTagsStep,
        createRowPeripheryTagsStep, createSubMatrixTagsStep,
        columnPeripheryStep, rowPeripheryStep,
        subMatrixStep, pivotStep)

      val numTiles: Int = n / b
      val tileSize: Int = b

      val begin: Long = java.lang.System.currentTimeMillis()
      val init: Runnable = new Runnable {
        def run: Unit = {
          graph.numTiles.put(Point.ZERO_1D, numTiles)
          graph.tileSize.put(Point.ZERO_1D, tileSize)

          var A_i: Int = 0
          var A_j: Int = 0
          var T_i: Int = 0
          var T_j: Int = 0

          var i = 0
          while (i < numTiles) {

            var j = 0
            while (j < numTiles) {

              // Allocate memory for the tiles.
              val temp: Array[Array[Double]] = Array.ofDim[Double](tileSize, tileSize)

              // Split the matrix into tiles and write it into the item space at time 0.
              // The tiles are indexed by tile indices (which are tag values).
              A_i = i * b
              T_i = 0
              while (T_i < tileSize) {
                A_j = j * b
                T_j = 0
                while (T_j < tileSize) {
                  temp(T_i)(T_j) = A(A_i)(A_j)
                  A_j += 1;
                  T_j += 1;
                }
                A_i += 1;
                T_i += 1;
              }
              graph.tiledMatrix.put(Point.create(i, j, 0), temp)

              j += 1
            }
            i += 1
          }

          graph.wholeComputationTag.put(Point.ZERO_1D)
        }
      }
      graph.run(init)
      val end: Long = java.lang.System.currentTimeMillis()

      println("Execution Time: " + (end - begin) + " ms.")
      println("tiledMatrix.size: " + graph.tiledMatrix.size())

      var mat_out: Array[Array[Double]] = null
      var k: Int = 0
      val w: BufferedWriter = new BufferedWriter(new FileWriter("lu_output.txt"))

      var i = 0
      while (i < numTiles) {

        var rowInTile = 0
        while (rowInTile < tileSize) {

          var j = 0
          while (j < numTiles) {
            if (i <= j) {
              mat_out = graph.tiledMatrix.safeGet(Point.create(i, j, i + 1))
            } else {
              mat_out = graph.tiledMatrix.safeGet(Point.create(i, j, j + 1))
            }

            var columnInTile: Int = 0
            while (columnInTile < tileSize) {
              w.write(mat_out(rowInTile)(columnInTile) + " ")
              columnInTile += 1;
            }

            j += 1
          }
          w.newLine()

          rowInTile += 1
        }

        i += 1
      }
      w.flush()

    } catch {
      case ex: Exception =>
        ex.printStackTrace()
    }
  }

  // Read the matrix from the input file.
  // The numbers in the matrix are whitespace separated
  def matrix_init(A: Array[Array[Double]], n: Int, fname: String): Unit = {
    val r: Reader = new BufferedReader(new FileReader(fname))
    val stok: StreamTokenizer = new StreamTokenizer(r)
    stok.parseNumbers()
    stok.nextToken

    var i: Int = 0
    while (i < n) {
      var j: Int = 0
      while (j < n) {
        if (stok.ttype == StreamTokenizer.TT_EOF) {
          throw new EOFException("Matrix size incorrect")
        }
        if (stok.ttype != StreamTokenizer.TT_NUMBER) {
          throw new IOException("A non-number has been read from the file: " + stok.sval)
        }
        A(i)(j) = stok.nval
        stok.nextToken

        j += 1;
      }

      i += 1;
    }
  }
}
